#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "audio.h"
#include "audio_internal.h"
#include "deps/resampler/speex_resampler.h"

typedef struct _FilterResamplerContext {
    uint32_t source_sample_rate;
    AudioBuffer *outputBuffer;
    SpeexResamplerState *st;
} FilterResamplerContext;

AudioPipelineStatus filter_resample_push(AudioPipelineElement *self, AudioBuffer *buffer) {
    FilterResamplerContext *context = (FilterResamplerContext *)self->ctx;

    if (self->sample_rate == context->source_sample_rate) {
        // Short circuit if we do not have to resample
        return self->next->push(self->next, buffer);
    }

    if (context->st == NULL) {
        fprintf(stderr, "ERROR: Pipeline not linked!\n");
        return PipelineError;
    }
    
    uint32_t buf_sz = (self->sample_rate / context->source_sample_rate) * buffer->buf_size;
    
    
    // Realloc if buffer size mismatch
    if ((context->outputBuffer != NULL) && (context->outputBuffer->buf_size < buf_sz)) {
        free_audio_buffer(context->outputBuffer);
        context->outputBuffer = NULL;
    }
    if (context->outputBuffer == NULL) {
        // at first push allocate appropriately sized output buffer
        context->outputBuffer = alloc_audio_buffer(buf_sz);
        if (context->outputBuffer == NULL) {
            fprintf(stderr, "ERROR: No memory for output buffer of size %d\n", buf_sz);
            return PipelineError;
        }
    }
    
    // calculate sample counts that fit into the buffers
    spx_uint32_t in_sz = (buffer->buf_size / self->channels / 2);
    spx_uint32_t out_sz = (context->outputBuffer->buf_size / self->channels / 2);
    
    // run the resampler on the buffer
    int result = speex_resampler_process_interleaved_int(
        context->st,
        (spx_int16_t *)buffer->data,
        &in_sz,
        (spx_int16_t *)context->outputBuffer->data,
        &out_sz
    );
    
    if (result != RESAMPLER_ERR_SUCCESS) {
        fprintf(stderr, "ERROR: Error while resampling: %d\n", result);
        return PipelineError;
    }
    
    if (in_sz != (buffer->buf_size / self->channels / 2)) {
        fprintf(stderr, "WARN: Processed %d samples of %d!\n", in_sz, (buffer->buf_size / self->channels / 2));
    }
    if (out_sz != (context->outputBuffer->buf_size / self->channels / 2)) {
        fprintf(stderr, "WARN: Output only %d samples of %d!\n", out_sz, (context->outputBuffer->buf_size / self->channels / 2));
    }
    
    // run next element of the pipeline
    return self->next->push(self->next, context->outputBuffer);
}

AudioPipelineStatus filter_resample_link(AudioPipelineElement *self, AudioPipelineElement *source) {
    FilterResamplerContext *context = (FilterResamplerContext *)self->ctx;

    context->source_sample_rate = source->sample_rate;
    self->bits_per_sample = source->bits_per_sample;
    self->channels = source->channels;

    if (self->sample_rate == source->sample_rate) {
        // Sample rates match, this filter is a no-op
        source->next = self;
        return PipelineStopped;
    }
    
    if (source->bits_per_sample != 16) {
        fprintf(stderr, "ERROR: Can only resample 16bits/sample, got: %d\n", source->bits_per_sample);
        return PipelineError;
    }
   
    if (context->st != NULL) {
        free_audio_buffer(context->outputBuffer);
        context->outputBuffer = NULL;
        speex_resampler_destroy(context->st);
        fprintf(stderr, "INFO: Re-linking, source has: %d Hz, we have %d Hz\n", source->sample_rate, self->sample_rate);
    }
    
    if (source->sample_rate > 0) {
        int err = 0;
        context->st = speex_resampler_init(source->channels, source->sample_rate, self->sample_rate, 4, &err);
        if (context->st == NULL) {
            fprintf(stderr, "ERROR: Error initializing resampler: %d\n", err);
            return PipelineError;
        }
    }
    
    source->next = self;
    if (self->next != NULL) {
        // relinking after samplerate change
        return self->next->link(self->next, self);
    }
    return PipelineStopped;
}

char *filter_resample_describe(AudioPipelineElement *self) {
    return "speexdsp resampler";
}

void filter_resample_destroy(AudioPipelineElement *self) {
    FilterResamplerContext *context = (FilterResamplerContext *)self->ctx;
    if (context->st != NULL) {
        speex_resampler_destroy(context->st);
    }
    if (context->outputBuffer) {
        free(context->outputBuffer);
    }
    free(context);
    free(self);
}

AudioPipelineElement *audio_filter_resample(uint32_t output_sample_rate) {
    AudioPipelineElement *self = calloc(1, sizeof(AudioPipelineElement));
    FilterResamplerContext *context = calloc(1, sizeof(FilterResamplerContext));
    
    self->sample_rate = output_sample_rate;
    
    self->ctx = context;
    self->describe = filter_resample_describe;
    self->start = filter_start_nop;
    self->reset = filter_reset_nop;
    self->stop = filter_stop_nop;
    self->push = filter_resample_push;
    self->link = filter_resample_link;
    self->destroy = filter_resample_destroy;
    self->type = AudioElementFilter;
    
    return self;
}
